      page  55,120
;***********************************************************************
;* PROGRAM TO REDEFINE ARROW KEYS, NUM-PAD KEYS AND X KEY FOR QMODEM
;***********************************************************************
; Complete description in QMFIX.DOC.
;***********************************************************************
; This program works by inserting itself into the keyboard interrupt
; vector (Interrupt 9). Whenever a key is struck, the software comes to
; our routine first. If our keys are hit, our routine handles it. All other
; scan codes are sent directly to the old interrupt handler.
;Other keys may be switched by replacing the scan codes & hex codes with others.
; (For shiftable keys, the shift-test block of code must be included.)
;This version saves memory by moving its resident portion down into part of
; the Program Segment Prefix space and adjusting its addresses accordingly.
;It can unload itself with the /U switch if no other TSRs have been loaded in
; the meantime.  It uses its own first 32 bytes as a signature to tell if it
; is already loaded.  It will load larger or smaller versions of itself
; depending on a command-line switch.
;***********************************************************************

.186		;Includes an instruction valid only on 186 or higher processor
CSEG	SEGMENT
	ASSUME	CS:CSEG,DS:CSEG

	ORG	0100H		;All ".COM" programs start at address 0100H

; THIS IS THE MAIN ENTRY POINT FOR THE PROGRAM

START:	JMP	INITIALIZE	;Jump to routine that loads this program as a TSR

;Equates for shifting resident portion into Program Segment Prefix before installing
Start2 = 40h			;Final start of code (in PSP)
Repos = Start - Start2		;Distance to move code

;After our TSR is installed, the Interrupt 9 vector points here.
; This means that the software comes here after every keystroke.

INT_9:
	PUSHA			;Save Registers
	MOV	AH,2		;Select Get-Shift-Status function
	INT	16H
	MOV	DL,AL		;Get the bios shift flags into DL register

	IN	AL,60H	;Read keyboard data port

	CMP	AL,2DH	;Is it the "X" DOWN scan code?
	JNE	VTST		;If not, check video memory
	CALL	X1		;If so, buffer remapped keystrokes
	JMP	SHORT	CLN1	; and return from the interrupt service routine

; When a key other than X is struck, the program first checks video memory
;(24- to 50-line text mode is assumed) for the character [ in the status
;line position used by Qmodem TD in terminal mode.  If not found, another menu
;is assumed and the arrow keys (and num-pad keys if loaded as the extended
;version) are not remapped.

VTST:	PUSHA
	MOV	CX,26		;Number of lines to check (to cover up to 50-line mode)
	MOV	BX,3894	;Point BX to line 24, column 27 (3840 = col. 0)
	PUSH	DS
	MOV	AX,CS:VIDDRESS	;Point DS to video memory address determined
	MOV	DS,AX			; on startup
RPT:	CMP	DS:BYTE PTR[BX],"["	;Is character at the screen location "["?
	JE	HIT				; (for Qmodem TD 4.6 status line)
	CMP	DS:BYTE PTR[BX+2],"["	;Is "[" at next screen location?
	JNE	NEXT				; (for Qmmodem 5.0 status line)
HIT:	POP	DS
	POPA
	PUSH	ES
	CALL	SW		;If character "[" seen, remap arrow keys
	POP	ES
	JC	CURK		;If arrow keys not pressed, see if other keys wanted
	JMP   SHORT CH1	;If arrow keys pressed, stuff desired codes into buffer
				; and clean up after interrupt
CURK:	CMP	CS:CURFLAG,0	;Was program loaded as smaller version?
	JE	KEYCON	;If so, no more keys to check; pass to original INT 9
	CALL	CURKEYS	;If loaded as extended version, check more keys
	JNZ	KEYCON	;If those keys not pressed, pass to original INT 9
Ch1:	CALL	STUF1		;If any of them pressed, stuff desired codes into buffer
	JMP	SHORT CLN1	; and clean up after interrupt
NEXT:	ADD	BX,160	;If "[" not seen, point to next display line
	LOOP	RPT		; and repeat up to 26 times
	POP	DS
	POPA		;Not in screen where we remap keys; fall into original INT 9

;The program comes here if no keys were remapped. The software jumps
; to the old interrupt vector so the original code can process any keys
; we don't care about.

KEYCON:
	POPA
	JMP	DWORD PTR CS:KEYVEC	;Jump to the old interrupt vector

;This code cleans up after the interrupt by clearing the interrupt in the
; interrupt controller chip.

CLN1:	CLI				;Don't allow any interrupts
	MOV	AL,20H		;Clear the interrupt controller
	OUT	20H,AL		; by sending 20h to port 20h

	POPA
	IRET			;Return from this interrupt

; If "[" found, the following code remaps the arrow keys, enabling proper
;cursor control in terminal mode.  To enter Scrollback mode, hit Home to get
;to the menu, then up arrow; the arrow keys then resume their normal
;functions, as they do in any screen in which the [Home] label disappears
;from the status line

;***********************************************************************
;* This section of code looks for and handles specified key presses
;***********************************************************************

TABLE DB	4BH,4DH,50H,48H	;Arrow-key scan codes to look for (in descending
					; order of matching ASCII codes for D, C, B, A)
SW:
	CLD			;Make sure that we are incrementing	
	MOV	CX,4		;Check for 4 arrow key scan codes
	PUSH	CS		;Set ES to point to table of arrow key scan codes
	POP	ES
	MOV	DI,OFFSET CS:TABLE	;Point ES:DI to scan code table
	REPNE SCASB		;Scan table until byte checked matches that in AL
	JZ	CUR		;If arrow key, buffer ^[[ and following ASCII letter
	STC			;If not found in 4 tries, key pressed wasn't arrow;
	RET			; set carry flag to skip buffer-stuffing call and exit

CUR:	CMP	CX,0		;Was key found Up Arrow?
	JNE	RMAP		;If not, remap for cursor control
	TEST	DL,8		;Alt key down?
	JZ	RMAP		;If not, remap for cursor control
	MOV	CX,4800H	;If so, send Up Arrow code to get scrollback mode
	RET

RMAP:	PUSH	CX		;Save number showing loop in which arrow key found
	CALL	ESCP		;Since arrow key was found, buffer ^[[
	POP	CX		;Restore number of loop in which arrow key found
	ADD	CL,41H 	;Convert this to matching ASCII char. D, C, B, or A
	RET

;The program arrives here if the X key was pressed. We test the BIOS shift
; flags for whether Alt and Left Shift or Ctrl and Left Shift are pressed.
; If so, we make the keystroke buffer our special series of keys.
; If Shift or Ctrl is pressed, we treat as normal shifted X or Ctrl-X.
; If no shift keys are pressed, we treat as normal unshifted x.

X1:	MOV	CH,2DH	;Put X key's scan code in CH

	PUSH	DX
	AND	DL,1010b
	CMP	DL,1010b	;Alt and Left Shift both down?
	POP	DX
	JE	TARG		;If so, buffer multiple keystrokes
	PUSH	DX
	AND	DL,1100b
	CMP	DL,1100b	;Ctrl and Left Shift both down?
	POP	DX
	JE	TARG		;If so, buffer multiple keystrokes
	TEST	DL,3		;Either Shift key down?
	JNZ	SH1		;If so, put shifted X's code in CX
	TEST	DL,4		;Ctrl key down?
	JNZ	CTRL		;If so, put CTRL-X's code in CX
	TEST	DL,8		;Alt key down?
	JNZ	ALT		;If so, put ALT-X's code in CX
	TEST	DL,40H	;Caps Lock active?
	JNZ	SH		;If so, put shifted X's code in CX
	JMP	SHORT	UNSH	;If none down, put unshifted key's code in CX
SH1:	TEST	DL,40H	;Both shift and Caps Lock active?
	JZ	SH		;If not, put shifted key's code in CX
UNSH:	MOV	CL,"x"	;Put unshifted key's code in CL
	JMP	SHORT	STUF
SH:	MOV	CL,"X"	;Put shifted key's code in CL
	JMP	SHORT	STUF
CTRL:	MOV	CL,18H	;Put Ctrl-X's code in CL
	JMP	SHORT	STUF
ALT:	XOR	CL,CL		;Put Alt-X's code in CL
	JMP	SHORT STUF
TARG:	MOV	CX,2000H	;Put Alt-D's scan code in CH and ASCII code in CL
	CALL	STUF1		;Stuff into keyboard buffer
	MOV	CL," "	;Put Space's ASCII code in CL
	CALL	STUF1
	MOV	CX,11BH	;Put Esc's scan code in CH and ASCII code in CL
	CALL	STUF1
	MOV	CX,2D00H	;Put Alt-D's scan code in CH and ASCII code in CL
	CALL	STUF1
	MOV	CL,"y"	;Put y's ASCII code in CL
STUF: CALL	STUF1		;Call subroutine to stuff code into buffer
	RET			;When all codes buffered, go on

ESCP:	MOV	CX,1A1BH	;Put ['s scan code in CH
	CALL	STUF1		; and ASCII code for ESC in CL (buffers ^[ esc. char.)
	MOV	CL,"["	;Put ['s ASCII code in CL
	CALL	STUF1
	RET

;This routine stuffs a scan code and ASCII value into the keyboard
; buffer, after the scan code has been put in CH and the ASCII code in CL

STUF1:
MOV	AH,05H	;Prepare to call BIOS write-to-buffer function
	INT	16H		;Call it
	CMP	AL,0		;Test return code for whether keyboard buffer full
	JE	RTN		;If full, try again
	MOV	AH,0
	INT	16H
RTN:	RET

KEYVEC	DW	?,?	;These four locations hold the old Interrupt 9 vector
VIDDRESS	DW	0	;Video memory address determined on startup
CURFLAG	DB	0	;Flags whether extended form of program installed

ISRend1 = $ + 1		;End of resident code and data for smaller version

; The routine below will be loaded to remap certain of the num-pad keys for
;cursor movement & character deletion in the Pico editor, if the extended
;program is specified by command-line switch; if not, it will be discarded
;with the transient code during installation.

CURKEYS:
	CMP	AL,4CH	;Numpad 5 scan code?
	JNE	DEL		;If not, look for next key

	MOV	CL,1		;Put Ctrl-A's ASCII code in CL
	RET

DEL:	CMP	AL,53H	;Del key scan code?
	JNE	EN		;If not, look for next key

	MOV	CL,4		;Put Ctrl-D's ASCII code in CL
	RET

EN:	CMP	AL,4FH	;End key scan code?
	JE	EN1		;If so, check whether End or Ctrl-End
	RET

EN1:	TEST	DL,4		;Ctrl key down?
	JNZ	EN2		;If so, buffer Ctrl-End code
	MOV	CL,5		;Put Ctrl-E's ASCII code in CL
	RET
EN2:	MOV 	CX,7500H	;Put Ctrl-End's scan/ASCII codes in CX
	RET

ISRend2 = $ + 1		;End of resident code and data for larger version

;Routine to install this TSR into memory

UNLOAD_MSG	DB	13,10,"QMFIX can't be uninstalled.",13,10,7
		DB	"Uninstall resident programs in reverse order.",13,10,"$"

NOT_INST_MSG	DB	13,10,"Can't uninstall QMFIX--not yet installed!",13,10,7,"$"
INSTALLED_MSG	DB	13,10,"QMFIX already installed",13,10,7,"$"
ALLOCATE_MSG	DB	"Memory allocation error",13,10,7,"$"
INSTALL_MSG		DB	13,10,"QMFIX installed",13,10,"$"
UNINSTALL_MSG	DB	13,10,"QMFIX uninstalled",13,10,"$"
PFLAG			DB	0	;Flag that no screen output wanted

;--------------------------------------------------------------------;
; Search memory for a copy of our code, to see if already installed. ;
;--------------------------------------------------------------------;

INITIALIZE:	ASSUME	DS:CSEG

		MOV	AH,0FH	;Select get-current-video-mode function
		INT	10H		; to determine current video mode
		CMP	AL,7		;Video mode 7 (monochrome text)?
		JL	B800		;If lower mode, assume video memory address B800h
		JG	EXIT		;If higher, graphics mode; abort
		MOV	VIDDRESS,0B000H	;If 7, store mono video memory address
		JMP	SHORT	F
B800:		MOV	VIDDRESS,0B800H	;If below 7, store alternate address

F:		CALL	FIND		;Determine whether our signature's in memory

;-------------------------------------------------;
; Parse the command line for uninstall parameter. ;
;-------------------------------------------------;

PARSE:	CLD			;Clear the direction flag (string operations forward)
		MOV	SI,81H		;Point to command line.
SEARCH_SWITCH:
		LODSB				;Get a byte.
		CMP	AL,13			;Is it carriage return?
		JZ	OVER 			;If yes, done here.
		CMP	AL,"/"		;Is there a switch character?
		JNZ	SEARCH_SWITCH	;If no, keep looking.
WGET:		LODSB				;Else, get the switch character.
		AND	AL,5FH		;Capitalize.
		CMP	AL,"X"		;Is it "X"?
		JNE	DE1			;If not, go on
		INC	CURFLAG		;Flag to install larger version of program
		JMP	SHORT	WGET		;Check next character
DE1:		CMP	AL,"S"		;Is it "S"?
		JNE	DE2			;If not, go on
		INC	PFLAG			;Flag that no screen output wanted
		JMP	SHORT	WGET		;Check next character
DE2:		CMP	AL,"U"		;Is it uninstall?
		JE	UNINSTALL		;If so, uninstall; if not, try installing

OVER:		MOV	AX,ES			;Are we already installed?
		MOV	BX,CS
		CMP	AX,BX
		JZ	INSTALL		;If no, install.
		XOR	AL,AL			;Else, ERRORLEVEL of zero.
		MOV	DX,OFFSET INSTALLED_MSG
		CALL	PRINT_STRING_1
		JMP	SHORT EXIT		;Exit.

;-------------------------------------------------------------------;
; Exit.	Return ERRORLEVEL code 0 if successful, 1 if unsuccessful. ;
;-------------------------------------------------------------------;

MSG_EXIT:	MOV	AL,1			;ERRORLEVEL = 1.

EXIT:		MOV	AH,4CH		;Terminate.
		INT	21H

;--------------------------------;
; This is the install procedure. ;
;--------------------------------;
INSTALL:
	MOV	ES,DS:[2CH]	;Place environment segment addr. for DOS to read
	MOV	AH,49H	;Prepare to call DOS free-memory-block function
	INT	21H		;Call it to free environment memory block

	MOV	DX,OFFSET ALLOCATE_MSG
	JC	MSG_EXIT		;If error, exit with message.

;Code to move resident code & data to lower memory space in Program Segment Prefix
	MOV	AX,CS			;Subtract specified no. of 16-byte
	SUB	AX,12			; paragraphs from code segment value and
	MOV	ES,AX			; make that ES value
	MOV	SI,OFFSET START	;Copy resident code to same offsets in
	MOV	DI,SI			; new lower segment
	CMP	CURFLAG,1		;Install larger version?
	JE	SIZ2			;If so, move more code
	MOV	CX,ISRend1 - Start	;Number of bytes to move--smaller version
	JMP	SHORT	MOR
SIZ2:	MOV	CX,ISRend2 - Start	;Number of bytes to move--larger version
MOR:	CLD
	REP MOVSB		;Copy code and data
	PUSH	ES		;Set DS to
	POP	DS		; new segment's value

	NOT	DS:BYTE PTR START[1]	;Change signature string so it can only be
						; found in the installed (resident) copy

	MOV	AX,3509H	;Get vector of INT 9 (returned in BX:ES)
	INT	21H
	MOV	WORD PTR DS:KEYVEC,BX	;Save the address in 1st word (BX part)...
	MOV	WORD PTR DS:KEYVEC[2],ES	; and 2nd word (ES part) of variable
	MOV	DX,OFFSET INT_9	;Get addr. where keyboard int. handler will be
	MOV	AX,2509H	;Replace keyboard interrupt with our address
	INT	21H
	PUSH	CS
	POP	DS		;Restore original data segment

	MOV	DX,OFFSET INSTALL_MSG	;Prepare to display install message.
	CALL	PRINT_STRING	;Display message unless flagged otherwise

	CMP	CURFLAG,1	;Install larger version?
	JE	SIZ3		;If so, keep more code
	MOV	DX,(ISRend1-repos+15)/16	;# of paragraphs to keep--smaller v.
	JMP	SHORT	MOR1
SIZ3:	MOV	DX,(ISRend2-repos+15)/16	;# of paragraphs to keep--larger v.
MOR1:	MOV	AX,3100H	;Return error code of zero.
	INT	21H		;Terminate but stay resident.

;---------------------------------------------------;
; This subroutine uninstalls the resident program.  ;
;---------------------------------------------------;

UNINSTALL:	MOV	DX,OFFSET NOT_INST_MSG
		MOV	CX,ES
		MOV	BX,CS
		CMP	CX,BX			;Is segment vector same?
		JZ	UNINSTALL_END	;If yes, not installed; exit.

		MOV	DX,OFFSET UNLOAD_MSG	;Error message if INT 9 changed.
		MOV	AX,3509H		;Get keyboard interrupt.
		INT	21H
		CMP	BX,OFFSET INT_9	;Has it been hooked by another?
		JNZ	UNINSTALL_END	;If not, exit with error message.
		MOV	BX,ES
		CMP	BX,CX			;Is the segment vector same?
		JNZ	UNINSTALL_END	;If yes, exit with error message.

		PUSH	ES		;Save segment where resident copy is loaded
		MOV	AX,ES
		ADD	AX,12		;Restore ES address originally allocated
		MOV	ES,AX		; by DOS before code was copied to lower memory
		MOV	AH,49H	;Return memory to system pool.
		INT	21H
		MOV	DX,OFFSET ALLOCATE_MSG
		POP	ES		;Restore segment value needed to reset int. vector
		JC	UNINSTALL_END	;Display message if problem.

		MOV	DX,WORD PTR ES:KEYVEC	;Restore old INT 9.
		MOV	DS,WORD PTR ES:KEYVEC[2]
		MOV	AX,2509H
		INT	21H

		PUSH	CS
		POP	DS			;Point to our data.
		MOV	DX,OFFSET UNINSTALL_MSG	;Prepare to display uninstall message.
		CALL	PRINT_STRING	;Display message unless otherwise flagged
		NOT	ES:BYTE PTR START[1]	;Kill resident signature string template
		JMP	MSG_EXIT
UNINSTALL_END:
		CALL	PRINT_STRING_1	;Display error messages regardless of flag
		JMP	MSG_EXIT		;And exit.

PRINT_STRING:
		CMP	PFLAG,0	;Was no-screen-output switch in command line?
		JNE	NO_SHOW	;If so, don't print message to screen
PRINT_STRING_1:			;Display message
		MOV	AH,9		;Print string via DOS.
		INT	21H
NO_SHOW:	RET

FIND:	CLD				;All string operations forward.
	MOV	BX,OFFSET START[1]	;Point to start of code.
	NOT	BYTE PTR [BX]	;Change a byte so no false match
					; with a disk cache copy
	MOV	AX,CS			;Store our segment in AX
	MOV	DX,AX			;Start at our segment
NEXT_PARA:
	INC	DX			;Next paragraph.
	MOV	ES,DX
	CMP	DX,AX			;Is it our segment?
	JNE	N1			;If not, continue search
					;If we cycle back to start, no copy in memory
	NOT	BYTE PTR [BX]	;Kill signature string template so not in memory
	RET				;Not resident; go on with CS=ES
N1:	MOV	SI,BX			;Else, point to our signature
	MOV	DI,BX			; and offset of possible match
	MOV	CX,32			;Check 32 bytes for match
	REP	CMPSB
	JNZ	NEXT_PARA		;If no match, keep looking
					;If match, signature already in memory
	NOT	BYTE PTR [BX]	;Kill signature string template so not in memory
	RET				;Resident; go on with CS!=ES

CSEG	ENDS
	END	START
